# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2013 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `filestore.migration` module.
"""

from unittest import TestCase
import os
from os import path
import stat
import time
import json
import shutil

import dbase32
from dbase32 import rfc3548

import filestore
from filestore import migration, misc

from . import TempDir


class TestConstants(TestCase):
    def test_B32ALPHABET(self):
        self.assertIsInstance(migration.B32ALPHABET, str)
        self.assertEqual(len(migration.B32ALPHABET), 32)
        self.assertEqual(len(set(migration.B32ALPHABET)), 32)
        self.assertEqual(migration.B32ALPHABET,
            ''.join(sorted(migration.B32ALPHABET))
        )
        self.assertEqual(migration.B32ALPHABET,
            ''.join(sorted(rfc3548.B32_FORWARD))
        )

    def test_B32NAMES(self):
        self.assertIsInstance(migration.B32NAMES, tuple)
        self.assertEqual(len(migration.B32NAMES), 1024)
        self.assertEqual(len(set(migration.B32NAMES)), 1024)
        self.assertEqual(migration.B32NAMES,
            tuple(sorted(migration.B32NAMES))
        )
        for name in migration.B32NAMES:
            self.assertIsInstance(name, str)
            self.assertEqual(len(name), 2)
            self.assertTrue(name.isupper() or name.isnumeric())
            self.assertTrue(set(name).issubset(migration.B32ALPHABET))

    def test_NAMES_DIFF(self):
        self.assertIsInstance(migration.NAMES_DIFF, tuple)
        self.assertEqual(len(migration.NAMES_DIFF), 124)
        self.assertEqual(len(set(migration.NAMES_DIFF)), 124)
        self.assertEqual(migration.NAMES_DIFF,
            tuple(sorted(migration.NAMES_DIFF))
        )
        for name in migration.NAMES_DIFF:
            self.assertIsInstance(name, str)
            self.assertEqual(len(name), 2)
            self.assertTrue(name.isupper() or name.isnumeric())
            self.assertIn(name, migration.B32NAMES)
            self.assertNotIn(name, filestore.DB32NAMES)


class TestFunctions(TestCase):        
    def test_is_v0_files(self):
        v0 = TempDir()
        for name in migration.B32NAMES:
            v0.mkdir(name)
        self.assertIs(migration.is_v0_files(v0.dir), True)
        v1 = TempDir()
        for name in filestore.DB32NAMES:
            v1.mkdir(name)
        self.assertIs(migration.is_v0_files(v1.dir), False)
        tmp = TempDir()
        filestore.FileStore.create(tmp.dir)
        files = tmp.join('.dmedia', 'files')
        self.assertTrue(path.isdir(files))
        self.assertIs(migration.is_v0_files(files), False)

    def test_b32_to_db32(self):
        for i in range(100):
            data = os.urandom(15)
            b32 = rfc3548.b32enc(data)
            db32 = dbase32.db32enc(data)
            self.assertEqual(migration.b32_to_db32(b32), db32)

    def test_v0_doc_to_v1(self):
        doc = {
            "_id": "4RNPIS23VFWZIOZDTPNHHENA", 
            "copies": 1, 
            "plugin": "filestore", 
            "time": 1359461597.704514, 
            "type": "dmedia/store"
        }
        self.assertEqual(migration.v0_doc_to_v1(doc), {
            "_id": "VKGIBLTUO8PSBHS6MIGAA7G3", 
            "copies": 1, 
            "plugin": "filestore", 
            "time": 1359461597.704514, 
            "type": "dmedia/store"
        })


class TestMigration(TestCase):
    def check_dir(self, fullname):
        self.assertEqual(path.abspath(fullname), fullname)
        st = os.stat(fullname, follow_symlinks=False)
        self.assertTrue(stat.S_ISDIR(st.st_mode), fullname)
        self.assertEqual(stat.S_IMODE(st.st_mode), 0o777, fullname)

    def check_file(self, fullname):
        self.assertEqual(path.abspath(fullname), fullname)
        st = os.stat(fullname, follow_symlinks=False)
        self.assertTrue(stat.S_ISREG(st.st_mode), fullname)
        self.assertEqual(stat.S_IMODE(st.st_mode), 0o444, fullname)

    def test_init(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)
        self.assertEqual(m.parentdir, tmp.dir)
        self.assertEqual(m.basedir, tmp.join(filestore.DOTNAME))
        self.assertEqual(m.files, tmp.join(filestore.DOTNAME, 'files'))
        self.assertEqual(m.files0, tmp.join(filestore.DOTNAME, 'files0'))
        self.assertEqual(m.store, tmp.join(filestore.DOTNAME, 'store.json'))
        self.assertEqual(m.store0, tmp.join(filestore.DOTNAME, 'store0.json'))
        self.assertEqual(m.filestore, tmp.join(filestore.DOTNAME, 'filestore.json'))

    def test_needs_migration(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)
        self.assertIs(m.needs_migration(), False)
        tmp.mkdir(filestore.DOTNAME)
        self.assertFalse(path.exists(m.filestore))
        self.assertFalse(path.exists(m.files))
        self.assertIs(m.needs_migration(), False)
        self.check_file(m.filestore)
        self.check_dir(m.files)

        # When there was a V0 files/ layout:
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        m.create_v0_files()
        self.assertFalse(path.exists(m.filestore))
        self.assertFalse(path.exists(m.files0))
        self.assertIs(m.needs_migration(), True)
        self.check_file(m.filestore)
        self.check_dir(m.files0)
        self.check_dir(m.files)

        # Again, after "soft" migration was already done:
        self.assertIs(m.needs_migration(), True)
        self.check_file(m.filestore)
        self.check_dir(m.files0)
        self.check_dir(m.files)

        # Remove files0, one more test:
        shutil.rmtree(m.files0)
        self.assertIs(m.needs_migration(), False)
        self.check_file(m.filestore)
        self.assertFalse(path.exists(m.files0))
        self.check_dir(m.files)

    def test_build_v0_simulation(self):
        V0 = misc.load_data('V0')
        V1 = misc.load_data('V1')
        tmp = TempDir()
        m = migration.Migration(tmp.dir)
        doc = m.build_v0_simulation()
        self.assertIsInstance(doc, dict)
        old_id = doc['_id']

        # First run, before rehashing has been done:
        id_map1 = {}
        ch_map = {}
        for (v0_id, v1_id, v1_ch) in m:
            id_map1[v0_id] = v1_id
            self.assertIsInstance(v1_ch, filestore.ContentHash)
            self.assertEqual(v1_ch.id, v1_id)
            ch_map[v1_id] = v1_ch
        self.assertEqual(len(id_map1), 6)
        self.assertEqual(len(ch_map), 6)
        self.assertEqual(len(set(id_map1.values())), 6)
        for (key, v0_id) in V0['root_hashes'].items():
            self.assertEqual(id_map1[v0_id], V1['root_hashes'][key])

        # Second run, after ID mapping via our "broken" symlinks is in place:
        id_map2 = {}
        for (v0_id, v1_id, v1_ch) in m:
            id_map2[v0_id] = v1_id
            self.assertIsNone(v1_ch)
        self.assertEqual(id_map2, id_map1)

        # See if we get a valid V1 FileStore layout:
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(migration.b32_to_db32(old_id), fs.id)
        fs.check_layout()
        ids = [st.id for st in fs]
        self.assertEqual(sorted(V1['root_hashes'].values()), ids)
        self.assertEqual(sorted(ch_map.keys()), ids)
        for (_id, ch) in ch_map.items():
            self.assertEqual(fs.verify(_id), ch)

    def test_join(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)

        # Cannot have empty parts:
        with self.assertRaises(ValueError) as cm:
            m.join()
        self.assertEqual(str(cm.exception), 'Cannot have empty parts')

        part1 = dbase32.random_id()
        self.assertEqual(m.join(part1),
            tmp.join(filestore.DOTNAME, part1)
        )
        part2 = dbase32.random_id()
        self.assertEqual(m.join(part1, part2),
            tmp.join(filestore.DOTNAME, part1, part2)
        )
        part3 = dbase32.random_id()
        self.assertEqual(m.join(part1, part2, part3),
            tmp.join(filestore.DOTNAME, part1, part2, part3)
        )

    def test_mkdir(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)
        part1 = dbase32.random_id()
        part2 = dbase32.random_id()

        # It shouldn't create basedir
        with self.assertRaises(FileNotFoundError) as cm:
            m.mkdir(part1)
        self.assertEqual(cm.exception.filename,
            tmp.join(filestore.DOTNAME, part1)
        )
        tmp.mkdir(filestore.DOTNAME)

        # Cannot have empty parts:
        with self.assertRaises(ValueError) as cm:
            m.mkdir()
        self.assertEqual(str(cm.exception), 'Cannot have empty parts')

        # Test with 1st dir:
        d1 = m.mkdir(part1)
        self.assertEqual(d1, tmp.join(filestore.DOTNAME, part1))
        self.check_dir(d1)
        with self.assertRaises(FileExistsError) as cm:
            m.mkdir(part1)
        self.assertEqual(cm.exception.filename, d1)

        # Test with 2nd dir:
        d2 = m.mkdir(part1, part2)
        self.assertEqual(d2, tmp.join(filestore.DOTNAME, part1, part2))
        self.check_dir(d2)
        with self.assertRaises(FileExistsError) as cm:
            m.mkdir(part1, part2)
        self.assertEqual(cm.exception.filename, d2)

    def test_write(self):
        doc = {
            '_id': dbase32.random_id(),
            'time': time.time(),
            'type': 'dmedia/store',
            'plugin': 'filestore', 
            'copies': 1, 
        }
        text = json.dumps(doc)
        name = dbase32.random_id(10).lower() + '.json'
        tmp = TempDir()
        m = migration.Migration(tmp.dir)

        # It shouldn't create basedir
        with self.assertRaises(FileNotFoundError) as cm:
            m.write(name, text)
        self.assertEqual(cm.exception.filename,
            tmp.join(filestore.DOTNAME, name)
        )
        tmp.mkdir(filestore.DOTNAME)

        # Should work when file doesn't already exist:
        f = m.write(name, text)
        self.assertEqual(f, tmp.join(filestore.DOTNAME, name))
        self.check_file(f)
        self.assertEqual(open(f, 'r').read(), text)

        # But fail when the file does exists:
        with self.assertRaises(FileExistsError) as cm:
            m.write(name, 'different text')
        self.assertEqual(cm.exception.filename, f)
        self.check_file(f)
        self.assertEqual(open(f, 'r').read(), text)

    def test_load_v0_doc(self):
        doc = {
            '_id': rfc3548.random_id(),
            'time': time.time(),
            'type': 'dmedia/store',
            'plugin': 'filestore', 
            'copies': 1, 
        }
        doc0 = {
            '_id': rfc3548.random_id(),
            'time': time.time(),
            'type': 'dmedia/store',
            'plugin': 'filestore', 
            'copies': 1, 
        }

        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        self.assertIsNone(m.load_v0_doc())

        # Should load store.json if it is the only one presest:
        m.write('store.json', json.dumps(doc))
        self.assertEqual(m.load_v0_doc(), doc)

        # But should load store0.json if *both* are presest:
        m.write('store0.json', json.dumps(doc0))
        self.assertEqual(m.load_v0_doc(), doc0)

        # Likewise, should load store0.json if it is the only one presest:
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        self.assertIsNone(m.load_v0_doc())
        m.write('store0.json', json.dumps(doc0))
        self.assertEqual(m.load_v0_doc(), doc0)

    def test_get_v1_doc(self):
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)

        new1 = m.get_v1_doc()
        filestore.check_doc(new1)
        new2 = m.get_v1_doc()
        filestore.check_doc(new2)
        self.assertNotEqual(new1['_id'], new2['_id'])

        # store.json, not isdb32(old_id):
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        _id = 'Z' * 24  # Z is not in DB32ALPHABET
        ts = time.time()
        old = {'_id': _id, 'time': ts, 'copies': 1}
        m.write('store.json', json.dumps(old))
        new = m.get_v1_doc()
        filestore.check_doc(new)
        self.assertEqual(new, {
            '_id': 'S' * 24,
            'time': ts,
            'type': 'dmedia/store',
            'plugin': 'filestore',
            'copies': 1,
        })

        # store.json, old['migrated'] is True:
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        _id = 'A' * 24  # A is in both DB32ALPHABET and B32ALPHABET
        ts = time.time()
        old = {'_id': _id, 'time': ts, 'copies': 2, 'migrated': True}
        m.write('store.json', json.dumps(old))
        new = m.get_v1_doc()
        filestore.check_doc(new)
        self.assertEqual(new, {
            '_id': _id,
            'time': ts,
            'type': 'dmedia/store',
            'plugin': 'filestore',
            'copies': 2,
        })

        # store.json, not isb32(old_id):
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        _id = '9' * 24  # 9 is not in B32ALPHABET
        ts = time.time()
        old = {'_id': _id, 'time': ts, 'copies': 0}
        m.write('store.json', json.dumps(old))
        new = m.get_v1_doc()
        filestore.check_doc(new)
        self.assertEqual(new, {
            '_id': _id,
            'time': ts,
            'type': 'dmedia/store',
            'plugin': 'filestore',
            'copies': 0,
        })

        # store.json, ambigous last case:
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        _id = 'A' * 24  # A is in both DB32ALPHABET and B32ALPHABET
        ts = time.time()
        old = {'_id': _id, 'time': ts, 'copies': 1}
        m.write('store.json', json.dumps(old))
        new = m.get_v1_doc()
        filestore.check_doc(new)
        self.assertEqual(new, {
            '_id': '3' * 24,
            'time': ts,
            'type': 'dmedia/store',
            'plugin': 'filestore',
            'copies': 1,
        })

        # store0.json is always interpreted as a V0, Base32 doc:
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        v0_id = rfc3548.random_id()
        v1_id = migration.b32_to_db32(v0_id)
        self.assertNotEqual(v0_id, v1_id)
        ts = time.time()
        old = {'_id': v0_id, 'time': ts, 'copies': 1}
        m.write('store0.json', json.dumps(old))
        new = m.get_v1_doc()
        filestore.check_doc(new)
        self.assertEqual(new, {
            '_id': v1_id,
            'time': ts,
            'type': 'dmedia/store',
            'plugin': 'filestore',
            'copies': 1,
        })

        # store0.json always takes precedence over store.json:
        _id = '8' * 24  # 8 is not in B32ALPHABET
        old = {'_id': _id, 'time': 1234.5678, 'copies': 1}
        m.write('store.json', json.dumps(old))
        self.assertEqual(m.get_v1_doc(), new)

    def test_v0_path(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)
        for i in range(100):
            _id = rfc3548.random_id(30)
            self.assertEqual(m.v0_path(_id),
                tmp.join(filestore.DOTNAME, 'files0', _id[:2], _id[2:])
            )

    def test_v1_path(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)
        for i in range(100):
            _id = dbase32.random_id(30)
            self.assertEqual(m.v1_path(_id),
                tmp.join(filestore.DOTNAME, 'files', _id[:2], _id[2:])
            )

    def test_create_v0_files(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)

        # It shouldn't create basedir
        with self.assertRaises(FileNotFoundError) as cm:
            m.create_v0_files()
        self.assertEqual(cm.exception.filename,
            tmp.join(filestore.DOTNAME, 'files')
        )
        tmp.mkdir(filestore.DOTNAME)

        # Now it should work:
        self.assertIsNone(m.create_v0_files())
        files = tmp.join(filestore.DOTNAME, 'files')
        self.check_dir(files)
        for name in migration.B32NAMES:
            self.check_dir(path.join(files, name))
        self.assertEqual(sorted(os.listdir(files)), list(migration.B32NAMES))

        # Should fail if .dmedia/files/ already exists:
        with self.assertRaises(FileExistsError) as cm:
            m.create_v0_files()
        self.assertEqual(cm.exception.filename, files)

    def test_create_v1_files(self):
        tmp = TempDir()
        m = migration.Migration(tmp.dir)

        # It shouldn't create basedir
        with self.assertRaises(FileNotFoundError) as cm:
            m.create_v1_files()
        self.assertEqual(cm.exception.filename,
            tmp.join(filestore.DOTNAME, 'files')
        )
        tmp.mkdir(filestore.DOTNAME)

        # Now it should work:
        self.assertIsNone(m.create_v1_files())
        files = tmp.join(filestore.DOTNAME, 'files')
        self.check_dir(files)
        for name in filestore.DB32NAMES:
            self.check_dir(path.join(files, name))
        self.assertEqual(sorted(os.listdir(files)), list(filestore.DB32NAMES))

        # Should fail if .dmedia/files/ already exists:
        with self.assertRaises(FileExistsError) as cm:
            m.create_v1_files()
        self.assertEqual(cm.exception.filename, files)

    def test_move_v0_files_to_files0(self):
        tmp = TempDir()
        tmp.mkdir(filestore.DOTNAME)
        m = migration.Migration(tmp.dir)
        m.create_v0_files()
        self.assertIsNone(m.move_v0_files_to_files0())

        # Check files/
        files = tmp.join(filestore.DOTNAME, 'files')
        self.check_dir(files)
        for name in filestore.DB32NAMES:
            self.check_dir(path.join(files, name))
        self.assertEqual(sorted(os.listdir(files)), list(filestore.DB32NAMES))

        # Check files0/
        files0 = tmp.join(filestore.DOTNAME, 'files0')
        self.check_dir(files0)
        for name in migration.B32NAMES:
            self.check_dir(path.join(files0, name))
        self.assertEqual(sorted(os.listdir(files0)), list(migration.B32NAMES))

