# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for `filestore.misc` module.
"""

from unittest import TestCase
import os
from os import path
import json
from hashlib import md5
from skein import skein512

from dbase32 import db32enc, isdb32, random_id, RANDOM_B32LEN
from dbase32.rfc3548 import b32enc

from . import TempDir

import filestore
from filestore.protocols import MIN_LEAF_SIZE
from filestore import misc, protocols


class TestFunctions(TestCase):
    def test_tohex(self):
        for i in range(10):
            h = md5(os.urandom(16))
            self.assertEqual(
                misc.tohex(h.digest()),
                h.hexdigest()
            )
        for i in range(10):
            h = skein512(os.urandom(16), digest_bits=240)
            self.assertEqual(
                misc.tohex(h.digest()),
                h.hexdigest()
            )

    def test_load_data(self):
        pkg = path.dirname(path.abspath(filestore.__file__))
        self.assertTrue(path.isdir(pkg))
        self.assertEqual(misc.DATADIR, path.join(pkg, 'data'))
        self.assertTrue(path.isdir(misc.DATADIR))

        for name in ('V0', 'V1', 'V1_hex', 'MD5SUMS'):
            f = path.join(misc.DATADIR, name + '.json')
            self.assertTrue(path.isfile(f))
            self.assertFalse(path.islink(f))
            self.assertEqual(
                misc.load_data(name),
                json.load(open(f, 'r'))
            )

    def test_build_leaves(self):
        obj = misc.build_leaves(MIN_LEAF_SIZE)
        self.assertEqual(obj,
            {
                'A': b'A',
                'B': b'B' * (MIN_LEAF_SIZE - 1),
                'C': b'C' * MIN_LEAF_SIZE,
            }
        )
        self.assertEqual(len(obj['A']), 1)
        self.assertEqual(len(obj['B']), MIN_LEAF_SIZE - 1)
        self.assertEqual(len(obj['C']), MIN_LEAF_SIZE)

        obj = misc.build_leaves(2 * MIN_LEAF_SIZE)
        self.assertEqual(obj,
            {
                'A': b'A',
                'B': b'B' * (2 * MIN_LEAF_SIZE - 1),
                'C': b'C' * (2 * MIN_LEAF_SIZE),
            }
        )
        self.assertEqual(len(obj['A']), 1)
        self.assertEqual(len(obj['B']), 2 * MIN_LEAF_SIZE - 1)
        self.assertEqual(len(obj['C']), 2 * MIN_LEAF_SIZE)

    def test_write_test_vector_files(self):
        tmp = TempDir()
        leaves = misc.build_leaves(protocols.VERSION1.leaf_size)
        self.assertIsNone(misc.write_test_vector_files(tmp.dir))
        self.assertEqual(
            sorted(os.listdir(tmp.dir)),
            ['A', 'B', 'C', 'CA', 'CB', 'CC']
        )
        for (key, data) in leaves.items():
            self.assertEqual(
                open(tmp.join(key), 'rb').read(),
                data
            )
            self.assertEqual(
                open(tmp.join('C' + key), 'rb').read(),
                leaves['C'] + data
            )

    def test_import_vector_files(self):
        fs = misc.TempFileStore()
        result = misc.import_vector_files(fs)
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 6)
        for ch in result:
            self.assertIsInstance(ch, filestore.ContentHash)
            self.assertIsInstance(ch.id, str)
            self.assertTrue(isdb32(ch.id))
            self.assertEqual(len(ch.id), 48)
            self.assertIsInstance(ch.file_size, int)
            self.assertGreaterEqual(ch.file_size, 1)
            self.assertIsInstance(ch.leaf_hashes, bytes)
            self.assertGreaterEqual(len(ch.leaf_hashes), 30)
            self.assertEqual(len(ch.leaf_hashes) % 30, 0)
        self.assertEqual(result,
            tuple(fs.verify(st.id) for st in fs)
        )

    def test_build_vectors(self):
        leaf_size = 2 * 1024 * 1024
        proto = protocols.SkeinProtocol(leaf_size, 200)
        leaves = misc.build_leaves(leaf_size)

        A0 = proto.hash_leaf(0, leaves['A'])
        A1 = proto.hash_leaf(1, leaves['A'])
        B0 = proto.hash_leaf(0, leaves['B'])
        B1 = proto.hash_leaf(1, leaves['B'])
        C0 = proto.hash_leaf(0, leaves['C'])
        C1 = proto.hash_leaf(1, leaves['C'])

        del leaves

        A  = proto.hash_root(1,                 A0)
        B  = proto.hash_root(leaf_size - 1,     B0)
        C  = proto.hash_root(leaf_size,         C0)
        CA = proto.hash_root(leaf_size + 1,     C0 + A1)
        CB = proto.hash_root(2 * leaf_size - 1, C0 + B1)
        CC = proto.hash_root(2 * leaf_size,     C0 + C1)

        self.assertEqual(
            misc.build_vectors(proto),
            {
                'leaf_hashes': {
                    'A': [db32enc(A0), db32enc(A1)],
                    'B': [db32enc(B0), db32enc(B1)],
                    'C': [db32enc(C0), db32enc(C1)],
                },
                'root_hashes': {
                    'A': db32enc(A),
                    'B': db32enc(B),
                    'C': db32enc(C),
                    'CA': db32enc(CA),
                    'CB': db32enc(CB),
                    'CC': db32enc(CC),
                }
            }
        )

        self.assertEqual(
            misc.build_vectors(proto, encoder=b32enc),
            {
                'leaf_hashes': {
                    'A': [b32enc(A0), b32enc(A1)],
                    'B': [b32enc(B0), b32enc(B1)],
                    'C': [b32enc(C0), b32enc(C1)],
                },
                'root_hashes': {
                    'A': b32enc(A),
                    'B': b32enc(B),
                    'C': b32enc(C),
                    'CA': b32enc(CA),
                    'CB': b32enc(CB),
                    'CC': b32enc(CC),
                }
            }
        )

    def test_get_integers(self):
        self.assertEqual(
            misc.get_integers(100),
            [0, 1, 99, 100, 101, 199, 200]
        )
        self.assertEqual(
            misc.get_integers(500),
            [0, 1, 499, 500, 501, 999, 1000]
        )

    def test_build_md5sums(self):
        A = b'A'
        B = b'B' * 99
        C = b'C' * 100
        self.assertEqual(
            misc.build_md5sums(100),
            {
                'files': {
                    'A': md5(A).hexdigest(),
                    'B': md5(B).hexdigest(),
                    'C': md5(C).hexdigest(),
                    'CA': md5(C + A).hexdigest(),
                    'CB': md5(C + B).hexdigest(),
                    'CC': md5(C + C).hexdigest(),
                },
                'integers': {
                    '0': md5(b'0').hexdigest(),
                    '1': md5(b'1').hexdigest(),
                    '99': md5(b'99').hexdigest(),
                    '100': md5(b'100').hexdigest(),
                    '101': md5(b'101').hexdigest(),
                    '199': md5(b'199').hexdigest(),
                    '200': md5(b'200').hexdigest(),
                },
                'personalization': {
                    'leaf': md5(protocols.PERS_LEAF).hexdigest(),
                    'root': md5(protocols.PERS_ROOT).hexdigest(),
                },
            }   
        )

        A = b'A'
        B = b'B' * 499
        C = b'C' * 500
        self.assertEqual(
            misc.build_md5sums(500),
            {
                'files': {
                    'A': md5(A).hexdigest(),
                    'B': md5(B).hexdigest(),
                    'C': md5(C).hexdigest(),
                    'CA': md5(C + A).hexdigest(),
                    'CB': md5(C + B).hexdigest(),
                    'CC': md5(C + C).hexdigest(),
                },
                'integers': {
                    '0': md5(b'0').hexdigest(),
                    '1': md5(b'1').hexdigest(),
                    '499': md5(b'499').hexdigest(),
                    '500': md5(b'500').hexdigest(),
                    '501': md5(b'501').hexdigest(),
                    '999': md5(b'999').hexdigest(),
                    '1000': md5(b'1000').hexdigest(),
                },
                'personalization': {
                    'leaf': md5(protocols.PERS_LEAF).hexdigest(),
                    'root': md5(protocols.PERS_ROOT).hexdigest(),
                },
            }   
        )

        self.assertEqual(
            misc.build_md5sums(8 * 1024 * 1024),
            misc.load_data('MD5SUMS')
        )   


class TestTempFileStore(TestCase):
    def test_create(self):
        """
        Test the TempFileStore.create() classmethod.
        """
        fs = misc.TempFileStore.create()
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertIsInstance(fs, misc.TempFileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsInstance(fs.id, str)
        self.assertEqual(len(fs.id), RANDOM_B32LEN)
        self.assertTrue(isdb32(fs.id))
        self.assertIsInstance(fs.copies, int)
        self.assertEqual(fs.copies, 1)

        store_id = random_id()
        fs = misc.TempFileStore.create(store_id, 2)
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertIsInstance(fs, misc.TempFileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsInstance(fs.id, str)
        self.assertEqual(len(fs.id), RANDOM_B32LEN)
        self.assertTrue(isdb32(fs.id))
        self.assertEqual(fs.id, store_id)
        self.assertIsInstance(fs.copies, int)
        self.assertEqual(fs.copies, 2)

        label = 'Dmedia' + random_id(10).lower()
        fs = misc.TempFileStore.create(label=label)
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertIsInstance(fs, misc.TempFileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsInstance(fs.id, str)
        self.assertEqual(len(fs.id), RANDOM_B32LEN)
        self.assertTrue(isdb32(fs.id))
        self.assertIsInstance(fs.copies, int)
        self.assertEqual(fs.copies, 1)
        self.assertEqual(fs.doc['label'], label)

    def test_init(self):
        fs = misc.TempFileStore()
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsInstance(fs.id, str)
        self.assertEqual(len(fs.id), RANDOM_B32LEN)
        self.assertTrue(isdb32(fs.id))
        self.assertIsInstance(fs.copies, int)
        self.assertEqual(fs.copies, 1)

        store_id = random_id()
        fs = misc.TempFileStore(store_id, 2)
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsInstance(fs.id, str)
        self.assertEqual(len(fs.id), RANDOM_B32LEN)
        self.assertTrue(isdb32(fs.id))
        self.assertEqual(fs.id, store_id)
        self.assertIsInstance(fs.copies, int)
        self.assertEqual(fs.copies, 2)

        label = 'Dmedia' + random_id(10).lower()
        fs = misc.TempFileStore(label=label)
        self.assertIsInstance(fs, filestore.FileStore)
        self.assertTrue(fs.parentdir.startswith('/tmp/TempFileStore.'))
        self.assertIsInstance(fs.id, str)
        self.assertEqual(len(fs.id), RANDOM_B32LEN)
        self.assertTrue(isdb32(fs.id))
        self.assertIsInstance(fs.copies, int)
        self.assertEqual(fs.copies, 1)
        self.assertEqual(fs.doc['label'], label)

    def test_repr(self):
        # With auto-generated ID:
        fs = misc.TempFileStore()
        self.assertTrue(isdb32(fs.id) and len(fs.id) == 24)
        self.assertEqual(repr(fs),
            'TempFileStore({!r}, {!r})'.format(fs.parentdir, fs.id)
        )

        # With explicit ID:
        store_id = random_id()
        fs = misc.TempFileStore(store_id)
        self.assertEqual(fs.id, store_id)
        self.assertEqual(repr(fs),
            'TempFileStore({!r}, {!r})'.format(fs.parentdir, fs.id)
        )

    def test_del(self):
        fs = misc.TempFileStore()
        parentdir = fs.parentdir
        self.assertTrue(path.isdir(parentdir))
        fs = None
        self.assertFalse(path.exists(parentdir))
